
' =============================================================================
'  PROJECT - HBridge controller
' =============================================================================
'  Decode a Servo signal, and control an H-Bridge with two PWM outputs
'  - The servo signal timing is 500 to 2500 uS 
'  - The HBridge signals are two PWM with frequency = 16 KHz (up to 32 Khz) 
' ============================================================================= 

' =============================================================================
'  How to set the PID parameters is explained at the end of this file
' =============================================================================

' =============================================================================
'  There are three setting sections: Debug / Servo-Control / Open-Loop
'  Use only one section and comment the other two
' ==============================================================================
'  The meaning of each setting is explained at the end of this file)
' =============================================================================

' =============================================================================
'  DEBUG SETTINGS   
' =============================================================================
'$define ControlOutAsFloat
'$define LimitMaxDuty
'$define ContextSave
'$define SerialDebugEnabled	
''$define TestLoopTime	
''$define TestInterruptTime	

''$define GainPotEnabled	
'$define FeedbackEnabled	
'$define AdcOversampling	

'$define PID_ENABLED		
'$define EnableDerivative	
''$define TrapezoidalArea	
'$define AntiWindUp			
'$define CompensateDeadArea	
	
'Symbol OscFreqHz = 4.4
'Symbol OscGain = 30	
'Symbol DeadArea = 800
'Symbol ZeroArea = 40
'Symbol IntegralGain = 1
'Symbol DerivativeGain = 1.8

' -----------------------------------------------------------------------
' Parameters for :	  1	|  	2 |   3 |   4   (test servo-motors)
' -----------------------------------------------------------------------
'  OscFreqHz = 		4.5 | 8.5 | 1.2 | 4.4 
'  OscGain = 		  6	|   8 |   4 |  30	(disable the gain pot)
'  DeadArea = 		  8	| 450 |  80 | 800	(enable CompensateDeadArea)
'  ZeroArea = 		  0	|  22 |   0 |  40	(enable CompensateDeadArea)
'  IntegralGain = 	  1	|   1 |   1 |   1 
'  DerivativeGain =   1	|   1 |   1 | 1.8
' -----------------------------------------------------------------------


' =============================================================================
'  SERVO-CONTROL SETTINGS
'  You could try one of the four settings (created for our test servo-motors)
'  But each process (not only servo-motors) should have different parameters 
'  So you should add the specific settings for your servo-control
' =============================================================================
'  How to generate your settings is explained at the end of this file  
' =============================================================================	
'$define ControlOutAsFloat
'$define ContextSave
'$define FeedbackEnabled
'$define AdcOversampling
'$define PID_ENABLED
'$define EnableDerivative 
'$define AntiWindUp
'$define CompensateDeadArea		
'' -------------------------- servo 1 (motor demultiplied) with supply = 18 volt 
''Symbol OscFreqHz = 4.5
''Symbol OscGain = 8
''Symbol DeadArea = 16
''Symbol ZeroArea = 0
''Symbol IntegralGain = 1
''Symbol DerivativeGain = 1	
'' -------------------------- servo 2 (10 turns potentiometer) with supply = 9 volt 
''Symbol OscFreqHz = 8.5
''Symbol OscGain = 8
''Symbol DeadArea = 450
''Symbol ZeroArea = 22
''Symbol IntegralGain = 1
''Symbol DerivativeGain = 1	
'' -------------------------- servo 3 (linear potentiometer) with supply = 9 volt 
''Symbol OscFreqHz = 1.2
''Symbol OscGain = 4
''Symbol DeadArea = 80	
''Symbol ZeroArea = 0
''Symbol IntegralGain = 1
''Symbol DerivativeGain = 1	
'' -------------------------- servo 4 (10 turns pot + belt) with supply = 9 volt 
'Symbol OscFreqHz = 4.4
'Symbol OscGain = 30	
'Symbol DeadArea = 800
'Symbol ZeroArea = 40
'Symbol IntegralGain = 1
'Symbol DerivativeGain = 1.8


' =============================================================================
'  OPEN-LOOP SETTINGS
'  To set a motor speed and direction with the standard signal (500 to 2500 uS)
'  Useful for : WAVE - SMART SCREW-DRIVER with the Screwdriver V3 PCB
' =============================================================================
$define ControlOutAsFloat
$define ContextSave
$define CompensateDeadArea
Symbol DeadArea = 100	
Symbol ZeroArea = 15
	

Device = 12F1572	' the 1572 memory is enough to enable all the options
'Device = 12F1571	' for the SCREW-DRIVER (with Feedback and PID disabled)
			

' #############################################################################
'  END OF THE USER AREA
' #############################################################################


' ============================================================================= 
'  TIMING PARAMETERS
' ============================================================================= 
' ------------------------------------------------------------------------------
'  Changing the PwmPeriod you must change also MaxDuty and multiply the gain
'  The min PWM frequency (for debug) is 500 Hz (PwmPeriod = 64000)
'  With PwmPeriod = 1000 the PWM frequency is 32 KHz and the resolution 10 bit
'  With PwmPeriod = 2000 the PWM frequency is 16 KHz and the resolution 11 bit
' ------------------------------------------------------------------------------
Symbol PwmPeriod = 2000
Symbol MaxDuty = 1999 				 ' used only if LimitMaxDuty is enabled
Symbol ChangeDirectionPauseUS = 200  ' more than 62.5 uS (16 KHz period)


Symbol MidCounts = 1500

' ------------------------------------------------------------------------------
'  Tsampling is servo-signal repetition time (normally 16 to 20 mS)
' ------------------------------------------------------------------------------
Symbol Tsampling = 0.016 
		

' =============================================================================
'  PID PARAMETERS PRE-CALCULATED BY THE PREPROCESSOR
' =============================================================================
Dim gain As Float = 2

' ------------------------ Ziegler-Nichols coeffs (see the end of this file)
$ifdef PID_ENABLED
	$ifdef EnableDerivative
		Symbol Ki = IntegralGain * (1 / OscFreqHz) / 2 
		Symbol Kd = DerivativeGain * (1 / OscFreqHz) / 8
		Symbol KiTs = Ki * Tsampling
		Symbol KdTs = Kd / Tsampling
		gain = OscGain * 0.6
	$else
		Symbol Ki = IntegralGain * (1 / OscFreqHz) * 0.8
		Symbol KiTs = Ki * Tsampling
		gain = OscGain * 0.45
	$endif
$endif

	
' =============================================================================
'  PIN ASSIGNEMENTS
' =============================================================================
'  8 VSS = Power supply 3.3 or 5 volt
'  7 RA0 = Debug serial OUT					ICSP-DAT	(TX for bootloader)
'  6 RA1 = Gain potentiometer				ICSP-CK		(RX For Bootloader)
'  5 RA2 = Feedback potentiometer
'  4 RA3 = Input for Servo-Signal (500 To 2500 uS)		(RESET for bootloader)
'  3 RA4 = H-BRIDGE PWM2
'  2 RA5 = H-BRIDGE PWM1                                 
'  1 VDD = Power supply GND

' -----------------------------------------------------------------------------
'  Init pins
' -----------------------------------------------------------------------------
$define DebugOut    PORTA.0
$define GainAdc		PORTA.1
$define FeedbackAdc PORTA.2 
$define ServoIN     PORTA.3
$define PWM2_Pin    PORTA.4  
$define PWM1_Pin    PORTA.5 

APFCON = %00000011 ' PWM2 on RA4 / PWM1 on RA5                  
    
Low    PWM1_Pin		' set as output and low
Low    PWM2_Pin		' set as output and low
Input  FeedbackAdc
Input  GainAdc
Output DebugOut
Input  ServoIN

Symbol GainAdcPin	  = 1
Symbol FeedbackAdcPin = 2


' =============================================================================
'  Setup the serial out (debug with the serial terminal)
' =============================================================================
$ifdef SerialDebugEnabled 
    'Declare hRsOut_Pin   = PORTA.0  (not needed it is always PORTA.0)
    Declare Hserial_Baud = 56000 ' 19200/56000...115200/256000/500000/1000000
    Declare HSerial_Terminator = CRLF
    ' --------------------------------------------------------------
    ' Example - send values to serial terminal
    ' RsOutLn "Duty1 = ", Dec wDuty1, " Duty2 = ", Dec wDuty2  
    ' For flating vars use Dec2 / Dec4 / Dec5 (default is 3 decimals)
$endif 

' =============================================================================
'  Optimizer  (not working - program bytes not reduced - sometimes increased)
' =============================================================================
'Declare Optimiser_Level = 3	' Maximum optimisation
'Declare Dead_Code_Remove = On	' Squeeze the code further

' =============================================================================
'  Setup the fuses
' =============================================================================
Config1 FOSC_INTOSC, WDTE_OFF, PWRTE_OFF, MCLRE_OFF, CP_OFF, BOREN_OFF, CLKOUTEN_OFF
Config2 PLLEN_OFF, LVP_OFF, BORV_LO, WRT_OFF, STVREN_OFF, LPBOREN_OFF 

' -----------------------------------------------------------------------------
' Set the clock source to internal oscillator at 16 MHz
' -----------------------------------------------------------------------------
' SPLLEN = 0    - PLL(x4) disabled
' IRCF   = 1111 - Internal oscillator at 16 Mhz HF
' UNINPL.= 0    - Unimplemented bit
' SCS    = 10   - System Clock Source from Internal Oscillator
' -----------------------------------------------------------------------------
'OSCCON = %01111010
'Declare Xtal = 16

' -----------------------------------------------------------------------------
' Set the clock source to internal oscillator at 32 MHz
' -----------------------------------------------------------------------------
' SPLLEN = 1    - PLL(x4) enabled
' IRCF   = 1110 - Internal oscillator at 8 Mhz HF
' UNINPL.= 0    - Unimplemented bit
' SCS    = 00   - System Clock Source from PLL (see DataSheet Page 48) 
' -----------------------------------------------------------------------------
OSCCON = %11110000  ' <-- to go at 32 MHz SCS must be 00 
Declare Xtal = 32

' -----------------------------------------------------------------------------
'  Oscillator tuning  +/- 2%  (unused)
' -----------------------------------------------------------------------------
'011111 = +31 (Max frequency)
'000001 = +1
'000000 = Oscillator running at factory-calibrated frequency
'111111 = -1 
'100000 = -31 (Min frequency)
' -----------------------------------------------------------------------------
'OSCTUNE = %00000

' =============================================================================
'  Timer0 initialization (unused)
' =============================================================================
' OPTION_REG = %11011111 ' Timer 0 clock source = Fosc/4 and prescaler = 256 
' Timer 0 clock = 32 MHz / 4 / 256 = 31250 Hz  
' Timer 0 step time = 32 uS

' =============================================================================
'  Timer1 initialized at 1 MHz  (Fosc/4 and prescaler = 1:8)
'  This initialization at 1 Mhz is valid only if Fosc = 32 MHz
' =============================================================================
Symbol TMR1CS1    = T1CON.7 ' Timer1 clock source select bits
Symbol TMR1CS0    = T1CON.6 ' Timer1 clock source select bits

Symbol T1CKPS1    = T1CON.5 ' Timer1 input clock prescale select bits
Symbol T1CKPS0    = T1CON.4 ' Timer1 input clock prescale select bits

Symbol Unimp1     = T1CON.3 ' Unimplemented
Symbol NOT_T1SYNC = T1CON.2 ' Timer1 external clock input sync control
Symbol Unimp2     = T1CON.1 ' Unimplemented
Symbol TMR1ON     = T1CON.0 ' Timer1 ON

T1CKPS0 = 1		' Prescaler
T1CKPS1 = 1		' 11 = 1:8 / 10 = 1:4 / 01 = 1:2 / 00 = 1:1

'TMR1CS1 = 0	' Timer1 clock source select bits
'TMR1CS0 = 0	' 11 = LFINTOSC / 10 = T1CKI / 01 = Fosc / 00 = Fosc/4

'NOT_T1SYNC = 1	' Not sync
'TMR1ON = 1		' Timer1 ON

' =============================================================================
'  INIT INTERRUPTS
' =============================================================================
Symbol GIE   = INTCON.7 ' Global Interrupt Enable
Symbol PEIE  = INTCON.6 ' Peripheral Interrupt Enable
Symbol T0IE  = INTCON.5 ' TMR0 Overflow Interrupt Enable
Symbol INTE  = INTCON.4 ' External Interrupt Enable
Symbol IOCIE = INTCON.3 ' Interrupt-On-Change Enable
Symbol T0IF  = INTCON.2 ' TMR0 Overflow Interrupt Flag
Symbol INTF  = INTCON.1 ' External Interrupt Flag
Symbol IOCIF = INTCON.0 ' Interrupt-On-Change Interrupt Flag

INTCON = %10001000    ' Global interrupts enable + IOC enable

IOCAN.3 = 1		' Enable interrupt on negative edge on PORTA.3 
IOCAP.3 = 1		' Enable interrupt on positive edge on PORTA.3
INLVLA.3 = 1	' PORTA.3 input as Schmitt Trigger (not TTL) 

On_Hardware_Interrupt GoTo Interrupt_Handler


' =============================================================================
'  Macro To Set PHASE / DUTY / PERIOD For PWM1 / PWM2 / PWM3
' =============================================================================
Symbol wPWM1PH_SFR = PWM1PHL.Word
Symbol wPWM1PR_SFR = PWM1PRL.Word
Symbol wPWM1DC_SFR = PWM1DCL.Word

Symbol wPWM2PH_SFR = PWM2PHL.Word
Symbol wPWM2PR_SFR = PWM2PRL.Word
Symbol wPWM2DC_SFR = PWM2DCL.Word

Symbol wPWM3PH_SFR = PWM3PHL.Word
Symbol wPWM3PR_SFR = PWM3PRL.Word
Symbol wPWM3DC_SFR = PWM3DCL.Word

$define HPWM1_Phase(pPhase)   : wPWM1PH_SFR = pPhase  : PWM1LDCONbits_LDA = 1
$define HPWM2_Phase(pPhase)   : wPWM2PH_SFR = pPhase  : PWM2LDCONbits_LDA = 1
$define HPWM3_Phase(pPhase)   : wPWM3PH_SFR = pPhase  : PWM3LDCONbits_LDA = 1
$define HPWM1_Period(pPeriod) : wPWM1PR_SFR = pPeriod : PWM1LDCONbits_LDA = 1
$define HPWM2_Period(pPeriod) : wPWM2PR_SFR = pPeriod : PWM2LDCONbits_LDA = 1
$define HPWM3_Period(pPeriod) : wPWM3PR_SFR = pPeriod : PWM3LDCONbits_LDA = 1
$define HPWM1_Duty(pDuty)     : wPWM1DC_SFR = pDuty   : PWM1LDCONbits_LDA = 1
$define HPWM2_Duty(pDuty)     : wPWM2DC_SFR = pDuty   : PWM2LDCONbits_LDA = 1 
$define HPWM3_Duty(pDuty)     : wPWM3DC_SFR = pDuty   : PWM3LDCONbits_LDA = 1 

' ============================================================================= 
'  INIT PWM REGISTERS (must set before to enable PWM to avoid glitches)
' ============================================================================= 
ResetPwmRegisters()

' =============================================================================
'  SETUP and ENABLE - PWM1 and PWM2 at 32 MHz
'  PWMxCLKCON = %00000001 ' HFINTOSC clock (16 MHz), no prescaler
'  PWMxCLKCON = %00000000 ' Fosc clock (32 MHz), no prescaler
' =============================================================================
PWM1CLKCON = %00000000 ' HFINTOSC clock, no prescaler
PWM1CON    = %11000000 ' peripheral enable, out enable, polarity, mode

PWM2CLKCON = %00000000 ' HFINTOSC clock, no prescaler
PWM2CON    = %11000000 ' peripheral enable, out enable, polarity, mode


' ============================================================================= 
'  INIT ADC
' ============================================================================= 
' ADFM   = 0    - 0 = Left justified / 1 = Right justified
' ADCS   = 000  - 000 = Fosc/2
' UNINPL.= 0    - Unimplemented bit
' ADPREF = 00   - 00 = VRPOS is connected to VDD

ADCON1 = %10000000 ' Clock Fosc/2, Connected to VDD, Right justified


' ============================================================================= 
'  INIT SERVO-CONTROLLER VARS
' ============================================================================= 
Dim cmdCounts As Word
Dim fbkCounts As SWord = MidCounts
Dim adcRaw As Word
Dim direction As Bit
$ifdef ControlOutAsFloat
	Dim controlOut As Float
$else
	Dim controlOut As SWord
$endif

' ============================================================================= 
'  INIT PID VARS
' ============================================================================= 
$ifdef PID_ENABLED
 	Dim errNew As Float
	Dim errOld As Float = 0
	Dim integral As Float = 0
	$ifdef EnableDerivative
		Dim derivative As Float
	$endif  
$endif

        
' ============================================================================= 
'  MAIN LOOP  (used only to test the ServoCommand Signal and for debug)
' ============================================================================= 
Do    
   	' -------------------------------------------------------------------------
   	'  Change the DebugOut to test the loop time with a scope 
	' ------------------------------------------------------------------------- 
    $ifdef TestLoopTime
        DebugOut = 1 - DebugOut
        DelayUS 100
    $endif
   
   	' -------------------------------------------------------------------------
   	'  Test the ServoCommand Signal
	' ------------------------------------------------------------------------- 
 	'  If the ServoCommand Signal stops for more than 50 mS then stop the motor
	'  The timer1H is incremented every 256 uS, so 200 steps are 50 mS
   	' -------------------------------------------------------------------------
 	If TMR1H > 200 Then
 		ResetPwmRegisters()
 	EndIf

	' -------------------------------------------------------------------------
	'  Send data to Theremino Terminal and SignalScope
	' -------------------------------------------------------------------------
    $ifdef SerialDebugEnabled
    	' ---------------------------------------------------- every 5 mS
        DelayMS 5
		' ---------------------------------------------------- send debug data
		$ifdef PID_ENABLED
			' ----------------------------------------------------------------------
			'  The following line sends the main PID parameters to the serial
			'  (cmdCounts / fbkCounts / gain / integral / derivative / controlOut)
			' ----------------------------------------------------------------------
			'  Using Theremino-Terminal (V1.2 and up) and Theremino SignalScope
  			'  you could see the variable values in real time, this is useful
  			'  to see the integral and derivative work and trim the PID parameters
			' ----------------------------------------------------------------------
			HRSOutLn SDec cmdCounts, " ", _
					 Dec fbkCounts, " ", _
					 Dec1 gain, " ", _
					 Dec0 integral, " ", _
					 Dec0 derivative, " ", _
					 Dec0 controlOut, " ", _
					 "sTime/cTime/gain/integ/deriv/cOut"	    
		$else
			' ----------------------------------------------------------------------
			'  The following line sends the oscillation test parameters to serial
			'  (cmdCounts / fbkCounts / gain / controlOut)
			' ----------------------------------------------------------------------
			'  Using Theremino-Terminal (V1.2 and up) and Theremino SignalScope
  			'  you could see the variable values in real time.
  			'  With PID disabled and gain-potentiometer enabled you increase
	  		'  the gain up to the oscillation, and measure the OscFrequency and Gain
			' ----------------------------------------------------------------------
  			HRSOutLn SDec cmdCounts, " ", _
					 Dec fbkCounts, " ", _
					 Dec3 gain, " ", _
					 SDec controlOut, " ", _
					 "sTime/cTime/gain/controlOut"
		$endif
	$endif
Loop    

' ============================================================================= 
'  SUBROUTINE - Reset PWM registers 
' =============================================================================
Sub ResetPwmRegisters()
	wPWM1PH_SFR = 0
	wPWM2PH_SFR = 0
	wPWM1DC_SFR = 0
	wPWM2DC_SFR = 0
	wPWM1PR_SFR = PwmPeriod
	wPWM2PR_SFR = PwmPeriod
	PWM1LDCONbits_LDA = 1
	PWM2LDCONbits_LDA = 1	
EndSub
 
' ============================================================================= 
'  Interrupt Service Routine 
' =============================================================================
Interrupt_Handler:  
    ' ----------------------------------------------------- context save
	$if _defined (SerialDebugEnabled) Or _defined (TestLoopTime) Or _defined (ContextSave)    
		Context Save
	$endif  
	$ifdef TestInterruptTime      
	    DebugOut = 1           
	$endif  
    ' ----------------------------------------------------- if ServoIN pin has changed
    ' (commented because we do not enable other interrupts)
    'If IOCIF = 1 Then
        ' ------------------------------------------------- test if the pin is HIGH    
        If ServoIN = 1 Then
            ' --------------------------------------------- reset and start the timer1
            TMR1H = 0
            TMR1L = 0  
            TMR1ON = 1 
        Else     
            ' --------------------------------------------- stop timer1 and read the time                               
            TMR1ON = 0
            cmdCounts = TMR1H << 8
            cmdCounts = cmdCounts + TMR1L
            TMR1ON = 1 
            ' -----------------------------------------------------------------
            '  The timer1 is restarted to allow the ServoSignal stop detection
            ' -----------------------------------------------------------------
            '  Now the cmdCounts has values from 500 to 2500
            '  (the timer1 clock is 1 MHz) (32 MHz Fosc/4 and prescaler 1:8)
            ' -----------------------------------------------------------------
			
            ' --------------------------------------------- read the gain ADC
            $ifdef GainPotEnabled       
            	gain = (ADIn GainAdcPin) / 8 + 1
            	' ----------------------------------------- now gain is 1.0 to 128.0
		    $endif
		    
		    ' --------------------------------------------- read the feedback ADC
		    $ifdef FeedbackEnabled
	            $ifdef AdcOversampling
	            	adcRaw = 0 
	              	Dim i As Byte
		            For i = 1 To 16
		            	adcRaw = adcRaw + ADIn FeedbackAdcPin
		            Next
					fbkCounts = 500 + adcRaw / 8
				$else
					adcRaw = ADIn FeedbackAdcPin
					fbkCounts = 500 + adcRaw * 2
				$endif
	            ' ----------------------------------------- now fbkCounts is 500 to 2500 uS
	        $else
         		fbkCounts = MidCounts   
		    $endif
		    
		    ' --------------------------------------------- PID
		    $ifdef PID_ENABLED
		    	errNew = cmdCounts - fbkCounts
		     	' ----------------------------------------- the gain multiplies P and also I and D 
		     	errNew = errNew * gain
		     	'
		     	$ifdef AntiWindUp
					Symbol s1 = errNew.Byte1
     				Symbol s2 = errOld.Byte1
         			If s1.7 <> s2.7 Then integral = 0  
				$endif
				'
				$ifdef TrapezoidalArea
			    	' ------------------------------------- calculate the trapezoidal area
			    	Symbol kiTrapezoidal = Ki / 2
					integral = integral + (errNew + errOld) * kiTrapezoidal
				$else
					' ------------------------------------- calculate the rectangular area
					integral = integral + errNew * KiTs
				$endif
				'
				' ----------------------------------------- add the integral to output
				controlOut = errNew + integral
				'
				' ----------------------------------------- add the derivative term
				$ifdef EnableDerivative  
					derivative = (errNew - errOld) * KdTs 
					controlOut = controlOut + derivative
				$endif 
			    '
				' ----------------------------------------- set errOld  
				errOld = errNew    
			$else
				' ----------------------------------------- proportional only (if PID is disabled)
				controlOut = cmdCounts - fbkCounts
				controlOut = controlOut * gain 
			$endif
			                      
            ' --------------------------------------------- test the direction 
            If controlOut < 0 Then 
                ' ----------------------------------------- if changing direction
                If direction <> 0 Then
                	direction = 0
                	' ------------------------------------- stop the old-direction PWM 
                	HPWM1_Duty(0)
                	' ------------------------------------- pause more than the PWM cycle time
                    DelayUS ChangeDirectionPauseUS
                EndIf  
                ' ----------------------------------------- dead area compensation
				$ifdef CompensateDeadArea 
					If controlOut < -ZeroArea Then
                		controlOut = controlOut - DeadArea
   					EndIf
                $endif 
               	' ----------------------------------------- limit the control to max PWM duty
               	$ifdef LimitMaxDuty
               		If controlOut < -MaxDuty Then 
						controlOut = -MaxDuty
					EndIf
				$endif
				' ----------------------------------------- Set the new value for the PWM
                HPWM2_Duty(0 - controlOut)
			Else
 				' ----------------------------------------- if changing direction
                If direction <> 1 Then
                	direction = 1
                	' ------------------------------------- stop the old-direction PWM 
                	HPWM2_Duty(0)
                	' ------------------------------------- pause more than the PWM cycle time
                    DelayUS ChangeDirectionPauseUS
                EndIf   
                ' ----------------------------------------- dead area compensation
                $ifdef CompensateDeadArea  
                	If controlOut > ZeroArea Then
                		controlOut = controlOut + DeadArea
               		EndIf
                $endif
                ' ----------------------------------------- limit the control to max PWM duty
                $ifdef LimitMaxDuty
	                If controlOut > MaxDuty Then 
						controlOut = MaxDuty
					EndIf
				$endif
				' ----------------------------------------- Set the new value For the PWM
                HPWM1_Duty(controlOut)
            EndIf                                                                                                              
        EndIf
        ' ------------------------------------------------- reset IOC flags for all the pins
        IOCAF = %000000  
    'End If  
    $ifdef TestInterruptTime  
        DebugOut = 0
    $endif 
$if _defined (SerialDebugEnabled) Or _defined (TestLoopTime) Or _defined (ContextSave)
	Context Restore
$else    
	Retfie
$endif


' ============================================================================= 
'  TIMINGS
' ============================================================================= 
' Read ADC with ADIn 			  			=  80 uS
' gain = 1.0 + adcRaw / 10000.0   			= 110 uS
' interrupt time            				= 190 uS 
' interrupt time with PID   				= 800 uS 
' interrupt time with PID and oversampling 	= 2 mS  (no problems up to 10 mS and more)

' PWM cycle times are 63 uS (16 KHz) or 2 mS (500 Hz)
' When changing direction we make a pause of 200 uS or 6 mS



' ============================================================================= 
'  HELP FOR THE TERMS USED
' ============================================================================= 

' -----------------------------------------------------------------------------
'  ContextSave
' -----------------------------------------------------------------------------
'  If some work is done in the main Loop the ContextSave is absolutely required 
'  (for example it must be enabled if sending data to the serial)
'  If the program memory is sufficient, it is always better to enable this flag
'  Otherwise it is possible to have sporadic glitches on the motor signal
 
' -----------------------------------------------------------------------------
'  SerialDebugEnabled
'  TestLoopTime
'  TestInterruptTime
' -----------------------------------------------------------------------------
'  Used to send serial data or pulses to the debug pin
'  To measure pulses time you could use a scope or the Theremino DAA
'  Or you could send data to the Slots with the application Theremino Terminal 

' -----------------------------------------------------------------------------
'  GainPotEnabled
' -----------------------------------------------------------------------------
' If enabled the gain can be set with a trimmer on the board
' The trimmer could be also used to set the motor speed (PID not required)
' If disabled the gain is defined in the Ziegler-Nichols parameters

' -----------------------------------------------------------------------------
'  FeedbackEnabled
' -----------------------------------------------------------------------------
' If enabled the firmware controls the position (like a normal Servo-Motor)
' If disabled the firmware controls the rotation speed (like a multiturn servo)

' -----------------------------------------------------------------------------
'  AdcOversampling
' -----------------------------------------------------------------------------
' The oversampling (by 16) reduces the feedback ADC noise   

' -----------------------------------------------------------------------------
'  Gain
' -----------------------------------------------------------------------------
'  In this implementation the gain multiplies also integral and deriv. terms
'  So the gain is not only the usual Kp but multiplies also Ki and Kd

' -----------------------------------------------------------------------------
'  PID_ENABLED
' -----------------------------------------------------------------------------
'  This enables Proportional-Integral-Derivative algorithm 
'  If PID is disabled the control is proportional only
'  Without PID the speed and the stability are reduced

' -----------------------------------------------------------------------------
'  EnableDerivative (it is a PID section and works only if PID is enabled)
' -----------------------------------------------------------------------------
'  Enabling this the control is a complete PID, otherwise it is a PI only

' -----------------------------------------------------------------------------
'  TrapezoidalArea (it is a PID section and works only if PID is enabled)
' -----------------------------------------------------------------------------
'  Normally the integral is calculated with a rectangular rule 
'  Enabling this flag a trapezioidal approximation is used
'  Teoretically the trapezioidal is more precise but:
'   - The data produced are delayed by the sampling time divided by two
'   - The program words are increased
'   - We have not observed improvements 

' -----------------------------------------------------------------------------
'  AntiWindUp (it is a PID section and works only if PID is enabled)
' -----------------------------------------------------------------------------
'  If the error sign is changed the AntiWindUp resets the integral value
'  This is best than the normal WindUp protection
'  The overshoot caused by the integral is completely eliminated

' -----------------------------------------------------------------------------
'  CompensateDeadArea
' -----------------------------------------------------------------------------
'  The dead area compensation can increase the precision and the stability
'  It could also be used in open loop systems to reduce the dead area

' -----------------------------------------------------------------------------
'  LimitMaxDuty
' -----------------------------------------------------------------------------
'  Normally not required because the PWM registers saturates if duty > period
'  It could be enabled for microcontrollers with different PWM behavior
'  Or it could enabled to protect from overflow if using extremely high gains
'  (the margin is 32 times the normal values ranging from -2000 to +2000)
'  Or it could be enabled while debugging to limit values to the effective area

' -----------------------------------------------------------------------------
'  ControlOutAsFloat
' -----------------------------------------------------------------------------
'  The control output can be SWord or Float 
'  - The SWord version saves about 300 program bytes
'  - The Float version does not risk the overflow even with very high gains
'    (an overflow status could rotate the motor in the wrong direction)

' -----------------------------------------------------------------------------
'  OscFreqHz
'  OscGain
'  DeadArea
'  ZeroArea
' -----------------------------------------------------------------------------
'  Those parameters are used to tune the PID with the Ziegler-Nichols method 
'  The Ziegler-Nichols method has been simplified and the PID improved
'  You must specify the Oscillation Frequency only, instead of Kp,Ki,Kd,Tu,Pu..
'  The stability and precision is increased with the Dead Area  
'  The settling time is reduced with an improved anti WindUp method


' =============================================================================
'  How to set the PID parameters
' =============================================================================
'  1) Disable the PID and increase the gain until the system oscillates 
'  2) Measure the oscillation frequency and the minimum gain to oscillate
'  3) Set the frequency in the PID parameters
'  4) Set the gain in the PID parameters
' -----------------------------------------------------------------------------
'  5) Test the movement
' -----------------------------------------------------------------------------
'  6) Set ZeroArea to 0
'  7) Increase DeadArea until the system oscillate
'  8) Decrease DeadArea by a 20%
'  9) Set ZeroArea = DeadArea / 20 (if ZeroArea is less than 10 set it to zero)
' -----------------------------------------------------------------------------
'  10) Test the movement
'  11) Increase or decrease IntegralGain And DerivativeGain as desired
' -----------------------------------------------------------------------------
'  Normally IntegralGain and DerivativeGain are "1" (Ziegler-Nichols default)
'  - Increasing IntegralGain reduces the settling time, but decrease stability
'  - Increasing DerivativeGain reduces the overshoot, but increase the noise
' =============================================================================




                

